from colouration import Colour


DEFAULT_COLOUR = Colour(hexadecimal='#AAAAAA')
DEFAULT_EDGE_TEXT_SIZE = 8
DEFAULT_ARROW_SIZE = 0.5
DEFAULT_FONT = 'helvetica'
DEFAULT_NODE_COLOUR_DIFFERENCE_RATIO = 0.5
DEFAULT_EDGE_OPACITY = 1
DEFAULT_EDGE_LABEL_STYLE = 'below'


class EdgeStyle:
	def __init__(
			self, colour=DEFAULT_COLOUR, opacity=DEFAULT_EDGE_OPACITY,
			font=DEFAULT_FONT, text_colour=None, text_size=DEFAULT_EDGE_TEXT_SIZE,
			arrow_size=DEFAULT_ARROW_SIZE, label_style=DEFAULT_EDGE_LABEL_STYLE,
			line_width=1
	):
		colour = colour or text_colour or DEFAULT_COLOUR
		colour = Colour(colour)
		self._colour = colour
		self._text_colour = None

		self._text_colour_based_on_main_colour = False

		self.text_colour = text_colour
		self.colour = colour

		self._opacity = opacity
		self._font = font
		self._text_size = text_size
		self._arrow_size = arrow_size
		self._label_style = label_style
		self._line_width = line_width

	def complement(self, dictionary):
		"""
		:type dictionary: dict
		"""
		for key, value in dictionary.items():
			if hasattr(self, key):
				setattr(self, key, value)
			elif not key.startswith('_'):
				setattr(self, f'_{key}', value)
			else:
				raise ValueError(f'{key} is not a valid attribute of EdgeStyle')

	def reset_colours(self):
		self._text_colour = None

	def copy(self):
		"""
		:rtype: EdgeStyle
		"""
		result = self.__class__(
			colour=self._colour, opacity=self._opacity,
			font=self._font, text_colour=self._text_colour, text_size=self._text_size,
			arrow_size=self._arrow_size, line_width=self._line_width
		)
		result._text_colour_based_on_main_colour = self._text_colour_based_on_main_colour
		return result

	def invert_colours(self):
		"""
		:rtype: EdgeStyle
		"""
		result = self.copy()

		if result._text_colour_based_on_main_colour:
			result._text_colour = None
		else:
			result.text_colour = result.text_colour.invert()

		if result.colour is not None:
			result.colour = result.colour.invert()

		return result

	@property
	def colour(self):
		"""
		:rtype: Colour
		"""
		return self._colour

	@colour.setter
	def colour(self, colour):
		if isinstance(colour, str):
			colour = Colour(obj=colour)
		self._colour = colour
		if self._text_colour is None:
			self._text_colour = colour
			self._text_colour_based_on_main_colour = True

	@property
	def text_colour(self):
		"""
		:rtype: Colour
		"""
		return self._text_colour or self.colour

	@text_colour.setter
	def text_colour(self, text_colour):
		if text_colour is None:
			if self.colour is not None:
				self._text_colour = self.colour
				self._text_colour_based_on_main_colour = True
		else:
			self._text_colour = Colour(text_colour)
			self._text_colour_based_on_main_colour = False

	def __getstate__(self):
		return {
			'_colour': self._colour,
			'_opacity': self._opacity,
			'_font': self._font,
			'_text_colour': self._text_colour,
			'_text_size': self._text_size,
			'_arrow_size': self._arrow_size,
			'_label_style': self._label_style,
			'_line_width': self._line_width
		}

	def __setstate__(self, state):
		self._colour = state['_colour']
		self._opacity = state['_opacity']
		self._font = state['_font']
		self._text_colour = state['_text_colour']
		self._text_size = state['_text_size']
		self._arrow_size = state['_arrow_size']
		self._label_style = state['_label_style']
		self._line_width = state['_line_width']

	@property
	def _graphviz_dictionary(self):
		return {
			'color': self.colour,
			'fontname': self._font,
			'fontcolor': self.text_colour,
			'fontsize': self._text_size,
			'arrowsize': self._arrow_size,
			'lblstyle': self._label_style,
			'penwidth': self._line_width
		}

	def get_graphviz_str(self):
		"""
		:rtype: str
		"""
		key_values = []
		for key, value in self._graphviz_dictionary.items():
			if isinstance(value, Colour):
				value = value.get_hexadecimal(opacity=self._opacity)

			if isinstance(key, str):
				key = key.strip('"')

			if isinstance(value, str):
				value = value.strip('"')

			if value is not None:
				key_values.append((key, value))

		return ' '.join([f'"{key}"="{value}"' for key, value in key_values])



from .Graph import Graph, draw_graph
from .Node import Node
from .Edge import Edge
from .get_ancestors import get_ancestors
from .get_descendants import get_descendants
from .graph_style.NodeStyle import NodeStyle


from ._GraphObj import GraphObj
from .graph_style.NodeStyle import NodeStyle


CORNER = u'\u2514'
TWO_WAY = u'\u251C'
HORIZONTAL = u'\u2500'
VERTICAL = u'\u2502'


class Node(GraphObj):
	def __init__(
			self, graph, name, value=None, label=None, style=None, index=0,
			**kwargs
	):
		"""
		:param .Graph.Graph graph: the graph this node belongs to
		:param str name: name of the node (code friendly, no weird characters, etc.) it should be unique in the graph
		:param value: any object stored in the node
		:param str label: a more presentable/user-friendly name
		:param style:
		:param kwargs:
		"""
		style = style or NodeStyle()

		super().__init__(
			graph=graph, id=name, value=value, label=label, style=style,
			# additional_styling_function=additional_styling_function,
			**kwargs
		)
		self._outward_edges_dict = dict()
		self._inward_edges_dict = dict()
		self._outward_edges_have_start_node = True
		self._inward_edges_have_end_node = True
		self._index = index

	@property
	def style(self):
		"""
		:rtype: NodeStyle
		"""
		return self._style

	@style.setter
	def style(self, style):
		"""
		:type style: NodeStyle or dict
		"""
		if isinstance(style, dict):
			style = NodeStyle(**style)
		self._style = style
		# self._style_is_native = style is not None

	@property
	def index(self):
		return self._index

	def __getstate__(self):
		state = super().__getstate__()
		state.update({
			'outward_edges_dict': self._outward_edges_dict,
			'inward_edges_dict': self._inward_edges_dict,
			'index': self._index
		})
		return state

	def __setstate__(self, state):
		super().__setstate__(state=state)
		self._outward_edges_dict = state['outward_edges_dict']
		self._inward_edges_dict = state['inward_edges_dict']
		self._outward_edges_have_start_node = False
		self._inward_edges_have_end_node = False
		self._index = state['index']

	def is_similar_to(self, other):
		"""
		:type other: Node
		:rtype: bool
		"""
		if not super().is_similar_to(other=other):
			return False
		if len(self._outward_edges_dict) != len(other._outward_edges_dict):
			return False
		if len(self._inward_edges_dict) != len(other._inward_edges_dict):
			return False
		for id, outward_edge in self._outward_edges_dict.items():
			other_outward_edge = other._outward_edges_dict[id]
			if not outward_edge.is_similar_to(other=other_outward_edge):
				return False
		for id, inward_edge in self._inward_edges_dict.items():
			other_inward_edge = other._inward_edges_dict[id]
			if not inward_edge.is_similar_to(other=other_inward_edge):
				return False
		return True

	def update_edges(self):
		if not self._outward_edges_have_start_node:
			for outward_edge in self._outward_edges_dict.values():
				outward_edge._start = self
				outward_edge._graph = self.graph
		self._outward_edges_have_start_node = True
		if not self._inward_edges_have_end_node:
			for inward_edge in self._inward_edges_dict.values():
				inward_edge._end = self
				inward_edge._graph = self.graph
		self._inward_edges_have_end_node = True

	@property
	def outward_edges_dict(self):
		return self._outward_edges_dict

	@property
	def inward_edges_dict(self):
		return self._inward_edges_dict

	@property
	def name(self):
		return self.id

	@property
	def outward_edge_ids(self):
		return list(self.outward_edges_dict.keys())

	@property
	def inward_edge_ids(self):
		return list(self.inward_edges_dict.keys())

	def get_outward_edge(self, id):
		return self.outward_edges_dict[id]

	def get_inward_edge(self, id):
		return self.inward_edges_dict[id]

	def append_outward_edge(self, edge):
		"""
		:type edge: GraphObj
		"""
		self.outward_edges_dict[edge.id] = edge

	def append_inward_edge(self, edge):
		"""
		:type edge: GraphObj
		"""
		self.inward_edges_dict[edge.id] = edge

	def remove_outward_edge(self, edge_id):
		del self.outward_edges_dict[edge_id]

	def remove_inward_edge(self, edge_id):
		del self.inward_edges_dict[edge_id]

	@property
	def label(self):
		if self._label:
			result = str(self.raw_label)
		else:
			result = str(self.id)
		return result.replace('"', '\\"')

	@label.setter
	def label(self, label):
		self._label = label

	def __str__(self):
		return f'Node:{self.id}'

	def __repr__(self):
		return str(self)

	def __lt__(self, other):
		"""
		:type other: Node
		:rtype: bool
		"""
		return (self.index, self.name) < (other.index, other.name)

	def __gt__(self, other):
		"""
		:type other: Node
		:rtype: bool
		"""
		return (self.index, self.name) > (other.index, other.name)

	def __le__(self, other):
		"""
		:type other: Node
		:rtype: bool
		"""
		return (self.index, self.name) <= (other.index, other.name)

	def __ge__(self, other):
		"""
		:type other: Node
		:rtype: bool
		"""
		return (self.index, self.name) >= (other.index, other.name)

	def __eq__(self, other):
		"""
		:type other: Node
		:rtype: bool
		"""
		return (self.index, self.name) == (other.index, other.name)

	def __ne__(self, other):
		"""
		:type other: Node
		:rtype: bool
		"""
		return (self.index, self.name) != (other.index, other.name)

	def get_tree_str(self, indentation='', already_added=None):
		"""
		:rtype: str
		"""
		if already_added is None:
			already_added = []

		tree_string = self.label

		if self in already_added:
			tree_string += '*\n'

		else:
			tree_string += '\n'
			num_children = len(self.children)

			for child_num, child in enumerate(self.children):

				already_added.append(self)
				if child_num+1 == num_children:
					tree_string += indentation + CORNER + ' ' + child.get_tree_str(
						indentation=indentation + '  ', already_added=already_added
					)

				else:
					tree_string += indentation + TWO_WAY + ' ' + child.get_tree_str(
						indentation=indentation + VERTICAL + ' ', already_added=already_added
					)

		return tree_string

	def get_graphviz_str(self):
		"""
		:rtype: str
		"""
		style = self.style

		if style is None:
			return f'"{self.id}" [label="{self.label}"]'

		else:
			return f'"{self.id}" [label="{self.label}" ' + style.get_graphviz_str() + ']'

	def connect_to(self, node, **kwargs):
		"""
		:type node: Node
		:rtype: GraphObj
		"""
		return self.graph.connect(start=self, end=node, **kwargs)

	def connect_from(self, node, **kwargs):
		"""
		:type node:
		:rtype:
		"""
		return self.graph.connect(start=node, end=self, **kwargs)

	@property
	def parents(self):
		"""
		:rtype: list[Node]
		"""
		return self.graph.get_parents(node=self)

	def has_parents(self):
		return self.graph.get_parents(node=self)

	def get_child_rank(self, parent):
		"""
		returns the rank of this node in terms of its index relative to its siblings, rank starts from 0
		:type parent: Node
		:rtype: int
		"""
		siblings = parent.children
		if self not in siblings:
			ValueError(f'{self} is not a child of {parent}')
		return sorted(siblings).index(self)

	@property
	def children(self):
		"""
		:rtype: list[Node]
		"""
		return self.graph.get_children(node=self)

	def has_children(self):
		return self.graph.has_children(node=self)

	@property
	def ancestors(self):
		"""
		:rtype: list[Node]
		"""
		return self.graph.get_ancestors(node=self, distance=False)

	@property
	def descendants(self):
		"""
		:rtype: list[Node]
		"""
		return self.graph.get_descendants(node=self, distance=False)

	@property
	def num_descendants(self):
		"""
		:rtype: int
		"""
		return len(self.descendants)

	@property
	def outward_edges(self):
		"""
		:rtype: list[Edge]
		"""
		result = []
		for edge in self._outward_edges_dict.values():
			if edge not in result:
				result.append(edge)
		return result

	@property
	def inward_edges(self):
		"""
		:rtype: list[Edge]
		"""
		result = []
		for edge in self._inward_edges_dict.values():
			if edge not in result:
				result.append(edge)
		return result

	@property
	def edges(self):
		"""
		:rtype: list[Edge]
		"""
		result = []
		for edge in self.outward_edges_dict.values():
			if edge not in result:
				result.append(edge)
		for edge in self.inward_edges_dict.values():
			if edge not in result:
				result.append(edge)
		return result

	def remove_edges(self):
		for edge in self.edges:
			edge.remove_self()

	@property
	def num_outward_edges(self):
		return len(self.outward_edges_dict)

	def is_in_loop(self):
		return self.graph.is_node_in_loop(node=self)

	def is_in_loop_with(self, other):
		return self.graph.are_nodes_in_same_loop(node1=self, node2=other)





def _get_descendants(obj, _objects_travelled=None, max_depth=None):
	"""
	:type _objects_travelled: list or NoneType
	:rtype: list[Node]
	"""
	if max_depth is not None:
		if max_depth <= 0:
			return [], {}
		else:
			max_depth -= 1
	_objects_travelled = _objects_travelled or []
	_objects_travelled.append(obj)
	children = obj.__children__()
	if len(children) == 0:
		return [], {}
	else:
		descendants = []
		descendants_dict = {1: []}
		for child in children:
			if child not in _objects_travelled:
				if child not in descendants:
					descendants.append(child)
					descendants_dict[1].append(child)
				child_descendants, child_descendants_dict = _get_descendants(
					obj=child, _objects_travelled=_objects_travelled, max_depth=max_depth
				)
				for descendant in child_descendants:
					if descendant not in descendants:
						descendants.append(descendant)
						for distance in child_descendants_dict.keys():
							descendants_dict[distance + 1] = child_descendants_dict[distance]
		return descendants, descendants_dict


def get_descendants(obj, distance=True, max_depth=None):
	descendants, descendants_dict = _get_descendants(obj=obj, _objects_travelled=None, max_depth=max_depth)
	if distance:
		return descendants_dict
	else:
		return descendants



from .StyleMaster import NodeStylist, EdgeStylist, StyleMaster
from .NodeStyle import NodeStyle
from .EdgeStyle import EdgeStyle
from .. import Node, Graph
from colouration import Colour
from .RootStylist import is_root_or_parents_are_in_loop


def get_style(node, pale_ratio, divergence_ratio, main_style):
	"""
	:type node: Node
	:rtype: NodeStyle
	"""
	if is_root_or_parents_are_in_loop(node):
		if main_style is None:
			return node.style.copy()
		else:
			return main_style.copy()
	else:
		if len(node.parents) == 1:
			parent = node.parents[0]
			num_siblings = len(parent.children)
			rank = node.get_child_rank(parent=parent)
			if main_style is None:
				style = get_style(
					node=parent, pale_ratio=pale_ratio, divergence_ratio=divergence_ratio, main_style=main_style
				)
			else:
				style = main_style
			if num_siblings == 0:
				colour = style.colour.copy(keep_id=True)
			else:
				colour = style.colour.increase_hue(
					amount=divergence_ratio * get_diverging_number(i=rank, n=num_siblings, reverse=False)
				)
			#colour = colour.pale(ratio=pale_ratio)
			colour.saturation = colour.saturation * 3 / 4
			new_style = style.copy()
			new_style.reset_colours()
			new_style.colour = colour
			return new_style
		else:
			styles = [
				get_style(node=parent, pale_ratio=pale_ratio, divergence_ratio=0, main_style=main_style)
				for parent in node.parents
				if not node.is_in_loop_with(parent)
			]
			colours = [style.colour for style in styles]
			if main_style is None:
				style = styles[0]
			else:
				style = main_style
			colour = Colour.mix(colours=colours).pale(ratio=pale_ratio)
			new_style = style.copy()
			new_style.reset_colours()
			new_style.colour = colour
			return new_style


def generate_diverging_numbers(n):
	if n == 1:
		return [0]
	elif n % 2 == 0:
		return [x/(n/2) for x in range(1, n // 2 + 1)] + [-x/(n/2) for x in range(1, n // 2 + 1)]
	else:
		return [0] + generate_diverging_numbers(n-1)


def get_diverging_number(i, n, reverse):
	return sorted(generate_diverging_numbers(n), reverse=reverse)[i]


class BranchNodeStylist(NodeStylist):
	def __init__(self, node_style=None, pale_ratio=0.05, divergence_ratio=0.05):
		self._divergence_ratio = divergence_ratio
		self._pale_ratio = pale_ratio
		self._node_style = node_style
		super().__init__(style=None, condition=None)

	def paint(self, graph):
		"""
		:type graph: Graph
		:rtype Graph
		"""

		for node in graph.nodes:
			if not is_root_or_parents_are_in_loop(node) and not node.is_in_loop():
				style = get_style(
					node=node, pale_ratio=self._pale_ratio, divergence_ratio=self._divergence_ratio,
					main_style=self._node_style
				)

				node.style = style
		return graph

	def copy(self):
		return self.__class__(pale_ratio=self._pale_ratio, divergence_ratio=self._divergence_ratio)


class BranchEdgeStylist(EdgeStylist):
	def __init__(self, edge_style=None):
		self._edge_style = edge_style
		super().__init__(style=None, condition=None)

	def paint(self, graph):
		"""
		:type graph: Graph
		:rtype Graph
		"""

		for edge in graph.edges:
			colour = edge.start.style.colour.darken(ratio=0.1)

			if self._edge_style is None:
				edge_style = EdgeStyle(colour=colour)
			else:
				edge_style = self._edge_style.copy()
				edge_style.reset_colours()
				edge_style.colour = colour

			if edge.value is not None:
				edge_style._arrow_size *= edge.value

			edge.style = edge_style
		return graph

	def copy(self):
		return self.__class__(edge_style=self._edge_style)


class BranchStylist(StyleMaster):
	def __init__(self, node_style=None, pale_ratio=0.05, divergence_ratio=0.05, edge_style=None):
		self._node_style = node_style
		self._pale_ratio = pale_ratio
		self._divergence_ratio = divergence_ratio
		self._edge_style = edge_style
		super().__init__(stylists=[
			BranchNodeStylist(node_style=node_style, pale_ratio=pale_ratio, divergence_ratio=divergence_ratio),
			BranchEdgeStylist(edge_style=edge_style)
		])

	def copy(self):
		return self.__class__(
			node_style=self._node_style, pale_ratio=self._pale_ratio, divergence_ratio=self._divergence_ratio,
			edge_style=self._edge_style
		)
